; AVR ASM bit-bang UART Tx routines
; https://preamp.org/static/avrasm-bitbang-uart-tx-routines/index.html
; 12/2023 by preamp.org

.nolist
.include "tn13def.inc" ; Define device ATtiny13 for simulator testing
;.include "m8adef.inc" ; Define device ATmega8A
.list

; not all registers are needed for all the routines
.def srg = r0
.def tmp = r16
.def buf = r17
.def cnt = r18
.def txbuf0 = r1
.def txbuf1 = r2
.def txbuf2 = r3
.def txbuf3 = r4
.def txbuf4 = r5
.def txbuf5 = r6
.def txbuf6 = r7
.def txbuf7 = r8
;.def txbuf8 = r29
;.def txbuf9 = r30

;Tx-port and pin
.equ txddr = DDRB
.equ txport = PORTB
.equ txpins = PINB
.equ txpin = 2        ; use 0..7

.cseg
.org 000000
main:

; init Tx-port
ldi tmp, (1<<txpin)
out txddr, tmp         ; set tx-pin to output
out txport, tmp        ; set high = idle = mark

; init stack pointer
.ifdef SPH
  ldi tmp, high(RAMEND)
  out SPH, tmp
.endif
ldi tmp, low(RAMEND)
out SPL, tmp

; simple loop generating some test calls
loop:
ldi tmp, 0x55
rcall sendusartfast2
ldi tmp, 0
rcall sendusartfast2
ldi tmp, 255
rcall sendusartfast2
nop
rjmp loop

;------------------------------------------------------------

sendusart:            ; 85/8
;bitbang tx routine, takes 8 clock cycles per baud, fastest it gets.
;byte to send in "tmp", need "buf", "cnt" (>=r16) and "srg" (r0..r31)
;the bit-to-be-sent is shifted into carry, and that flag
;copied into the buffer.
in buf, txpins
cbr buf, (1<<txpin)   ; start bit, space, 0
out txport, buf       ; PIN CHANGE
ldi cnt, 9
nop                   ; necessary nop
sec                   ; this will become our stop bit w/first ror!
sendusartdec:
ror tmp               ; rotate right through carry
in srg, SREG
bst srg, 0
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
dec cnt
brne sendusartdec
ret

;------------------------------------------------------------

sendusart2:           ; 77/7  /10  /14  /16
;bitbang tx routine, takes 7 clock cycles per baud, fastest it gets.
;byte to send in "tmp", need "buf", "cnt" (>=r16) and "srg" (r0..r31)
in buf, txpins
cbr buf, (1<<txpin)   ; start bit, space, 0
out txport, buf       ; PIN CHANGE
nop                   ; necessary nop
nop                   ; necessary nop
nop                   ; necessary nop
ldi cnt, 8
bst tmp, 0
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
sec                   ; this will become our stop bit w/first ror!
ror tmp               ; rotate right through carry
nop                   ; necessary nop
nop                   ; necessary nop
sendusartdec2:
bst tmp, 0
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
ror tmp               ; rotate right through carry
dec cnt
brne sendusartdec2
ret

;------------------------------------------------------------

sendusartplus:        ; 38/3 54/5  /7  /14  /16
;bitbang tx routine, takes 3 clock cycles per baud, fastest it gets.
;byte to send in "tmp", need "buf" (>=r16)
in buf, txpins
;cbi txport, (1<<txpin) ; PIN CHANGE; set start bit, space, 0
cbr buf, (1<<txpin)   ; start bit, space, 0
out txport, buf       ; PIN CHANGE
bst tmp, 0
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
bst tmp, 1
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
bst tmp, 2
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
bst tmp, 3
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
bst tmp, 4
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
bst tmp, 5
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
bst tmp, 6
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
bst tmp, 7
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
nop                   ; necessary nop
sbr buf, (1<<txpin)   ; stop bit, mark, 1
;--- additional delay here
out txport, buf       ; PIN CHANGE
ret


;------------------------------------------------------------

sendusartloop:        ; 123/12
;bitbang tx routine, adjustable clock cycles per baud (12 minimum).
;byte to send in "tmp", need "buf" (>=r16) and "srg" (r0..r31)
;the bit-to-be-sent is shifted into carry, and that flag
;copied into the buffer.
in buf, txpins
cbr buf, (1<<txpin)   ; start bit, space, 0
out txport, buf       ; PIN CHANGE
nop                   ; necessary nop
nop                   ; necessary nop
nop                   ; necessary nop
sec                   ; this will become our stop bit w/first ror!
;--- additional delay here
rcall sendusartwaitloop
;--- additional delay here
rcall sendusartwaitloop
;--- additional delay here
rcall sendusartwaitloop
;--- additional delay here
rcall sendusartwaitloop
;--- additional delay here
rcall sendusartwaitloop
;--- additional delay here
rcall sendusartwaitloop
;--- additional delay here
rcall sendusartwaitloop
;--- additional delay here
rcall sendusartwaitloop
;--- additional delay here
rcall sendusartwaitloop  ; stop bit
ret

sendusartwaitloop:
ror tmp               ; rotate right through carry
in srg, SREG
bst srg, 0
bld buf, txpin
;--- additional delay here
out txport, buf       ; PIN CHANGE
ret


;------------------------------------------------------------

sendusartfast:        ; 45/1  /2  /5  /7  /14  /16
; only 1 clock per baud, but 45 clocks per byte
; pre-charge buffer registers (can be <r16, except "buf")
in buf, txpins
bst tmp, 0
bld buf, txpin
mov txbuf0, buf
bst tmp, 1
bld buf, txpin
mov txbuf1, buf
bst tmp, 2
bld buf, txpin
mov txbuf2, buf
bst tmp, 3
bld buf, txpin
mov txbuf3, buf
bst tmp, 4
bld buf, txpin
mov txbuf4, buf
bst tmp, 5
bld buf, txpin
mov txbuf5, buf
bst tmp, 6
bld buf, txpin
mov txbuf6, buf
bst tmp, 7
bld buf, txpin
mov txbuf7, buf
sbr buf, (1<<txpin)    ; stop bit, mark, 1
; go!
cbi txport, (1<<txpin); PIN CHANGE; set start bit, space, 0
;--- additional delay here
out txport, txbuf0    ; PIN CHANGE
;--- additional delay here
out txport, txbuf1    ; PIN CHANGE
;--- additional delay here
out txport, txbuf2    ; PIN CHANGE
;--- additional delay here
out txport, txbuf3    ; PIN CHANGE
;--- additional delay here
out txport, txbuf4    ; PIN CHANGE
;--- additional delay here
out txport, txbuf5    ; PIN CHANGE
;--- additional delay here
out txport, txbuf6    ; PIN CHANGE
;--- additional delay here
out txport, txbuf7    ; PIN CHANGE
;--- additional delay here
out txport, buf       ; PIN CHANGE
ret

;------------------------------------------------------------

sendusartfast2:       ; /1  /2  /5  /7  /14  /16
; only 1 clock per baud, but 46 clocks per byte
; pre-charge buffer registers (can be <r16, except "buf")
in buf, txpins
bst tmp, 0
bld buf, txpin
mov txbuf0, buf
bst tmp, 1
bld buf, txpin
mov txbuf1, buf
bst tmp, 2
bld buf, txpin
mov txbuf2, buf
bst tmp, 3
bld buf, txpin
mov txbuf3, buf
bst tmp, 4
bld buf, txpin
mov txbuf4, buf
bst tmp, 5
bld buf, txpin
mov txbuf5, buf
bst tmp, 6
bld buf, txpin
mov txbuf6, buf
bst tmp, 7
bld buf, txpin
mov txbuf7, buf
mov tmp, buf
cbr tmp, (1<<txpin)    ; start bit, space, 0
sbr buf, (1<<txpin)    ; stop bit, mark, 1
; go!
out txport, tmp       ; PIN CHANGE
;--- additional delay here
out txport, txbuf0    ; PIN CHANGE
;--- additional delay here
out txport, txbuf1    ; PIN CHANGE
;--- additional delay here
out txport, txbuf2    ; PIN CHANGE
;--- additional delay here
out txport, txbuf3    ; PIN CHANGE
;--- additional delay here
out txport, txbuf4    ; PIN CHANGE
;--- additional delay here
out txport, txbuf5    ; PIN CHANGE
;--- additional delay here
out txport, txbuf6    ; PIN CHANGE
;--- additional delay here
out txport, txbuf7    ; PIN CHANGE
;--- additional delay here
out txport, buf       ; PIN CHANGE
ret


;------------------------------------------------------------

sendusartspecial:     ; 29/2  56/5  74/7  101/10  137/14  155/16
;this thing acts directly on the port with no buffering.
;only works with pin 0 as the output pin!
;the whole port will be written to, so set the remaining
;pins as inputs and don't use them for anything serious.
cbi txport, 0         ; PIN CHANGE; set start bit, space, 0
nop                   ; necessary nop
;--- additional delay here
out txport, tmp       ; PIN CHANGE
lsr tmp
;--- additional delay here
out txport, tmp       ; PIN CHANGE
lsr tmp
;--- additional delay here
out txport, tmp       ; PIN CHANGE
lsr tmp
;--- additional delay here
out txport, tmp       ; PIN CHANGE
lsr tmp
;--- additional delay here
out txport, tmp       ; PIN CHANGE
lsr tmp
;--- additional delay here
out txport, tmp       ; PIN CHANGE
lsr tmp
;--- additional delay here
out txport, tmp       ; PIN CHANGE
lsr tmp
;--- additional delay here
out txport, tmp       ; PIN CHANGE
;nop                   ; only for cores w/1PC for CBI/SBI
;--- additional delay here
sbi txport, 0         ; PIN CHANGE; set stop bit, mark, 1
ret

;------------------------------------------------------------

sendusartstack:       ; 70/2  /5  /7  /10  /14  /16
; pre-charge stack buffer in reverse order
in buf, txpins
bst tmp, 7
bld buf, txpin
push buf
bst tmp, 6
bld buf, txpin
push buf
bst tmp, 5
bld buf, txpin
push buf
bst tmp, 4
bld buf, txpin
push buf
bst tmp, 3
bld buf, txpin
push buf
bst tmp, 2
bld buf, txpin
push buf
bst tmp, 1
bld buf, txpin
push buf
bst tmp, 0
bld buf, txpin
push buf
cbr buf, (1<<txpin)    ; start bit, space, 0
; go!
out txport, buf        ; PIN CHANGE
;--- additional delay here
pop buf
out txport, buf        ; PIN CHANGE
;--- additional delay here
pop buf
out txport, buf        ; PIN CHANGE
;--- additional delay here
pop buf
out txport, buf        ; PIN CHANGE
;--- additional delay here
pop buf
out txport, buf        ; PIN CHANGE
;--- additional delay here
pop buf
out txport, buf        ; PIN CHANGE
;--- additional delay here
pop buf
out txport, buf        ; PIN CHANGE
;--- additional delay here
pop buf
out txport, buf        ; PIN CHANGE
;--- additional delay here
pop buf
out txport, buf        ; PIN CHANGE
;--- additional delay here
sbr buf, (1<<txpin)    ; stop bit, mark, 1
out txport, buf        ; PIN CHANGE
ret

;

